import keras
from keras import model, layers


def save_model(model, verbose=True, config == True):
    """Saves the model in the disk. We can both save the architecture in JSON file and model in .h5 file

    Parameters:
    model(object): Giving the reference of the model
    verbose(bool): Tells and prints information of the activity
    config(bool): It tells function to save architecture of the model or not

    """
  if config == True:
    model_json = model.to_json()
    with open("model.json", "w") as json_file:
        json_file.write(model_json)
    # serialize weights to HDF5
    # model.save_weights("model.h5")
    if verbose==True:
      print("Saved model architecture to disk")
  model.save("model.h5")

  
 
# later...
 
# load json and create model

def load_model(config=True):
    """
    Loads the model from memory to disk

    Parameters:
    config(bool): To load architecture of model or not

    Returns:
    reference of the loaded model
    """
  json_file = open('model.json', 'r')
  loaded_model_json = json_file.read()
  json_file.close()
  model = model_from_json(loaded_model_json)
  # load weights into new model
  model.load_weights("model.h5")
  return model



# -*- coding: utf-8 -*-
"""ml.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1-xpQSgFCkv7rMA0ZblqrgkOv7H3Xxv0a
"""

def ML(X, Y, size=0.25, *args):

  """
  This function is used to train different machine learning algorithm such as LogisticRegesstion, DecisionTree, Random Forest, svm etc..
  and prints its metrics.

  Parameters:
  X - input data
  Y - output vector
  size - validation or testing data percentage
  args - ml algorithms

  Returns:
  returns trained ml algorithms
  """


  from sklearn.metrics import accuracy_score, confusion_matrix, classification_report
  from sklearn.model_selection import train_test_split
  import pandas as pd
  import numpy
##creating a function to handle classification algorithms
  print('____________________________________________________')
  print(".....................Py-AutoML......................") 
  #print(args)  
  if  not args:
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.tree import DecisionTreeClassifier
    from sklearn.neighbors import KNeighborsClassifier
    from sklearn.linear_model import LogisticRegression
    from sklearn.svm import SVC

    args = (SVC(), RandomForestClassifier(), DecisionTreeClassifier(), KNeighborsClassifier(), LogisticRegression(max_iter = 7000))






  # dropping rows containing missing values
  X = pd.DataFrame(X)
  X.dropna(inplace= True, axis = 0)

  Y = pd.DataFrame(Y)
  Y.dropna(inplace= True, axis = 0)

  ##splitting the data into training and testing data and setting the test_size
  X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size=size, random_state = 0)   

  ##training the model with train data
  acc_score = []
  for alg in args:
      alg.fit(X_train, y_train.values.ravel())
      prediction = alg.predict(X_test)

      ##outputing statistics on the performance of the individual models
      ##statistics printed out are the accuracy score, confusion matrix and clasification report
      
      print('____________________________________________________')
      print(alg.__class__.__name__,"______________________________","\n")
      print("Accuracy Score for "+ alg.__class__.__name__+" is \n"+ str(accuracy_score(y_test, prediction)))
      print('\n')

      print("Confusion Matrix for "+ alg.__class__.__name__+" is \n"+ str(confusion_matrix(y_test, prediction)))
      print('\n')

      print("Classification Report for "+ alg.__class__.__name__+" is \n"+ str(classification_report(y_test, prediction)))
      print('\n')


      ##saving the accuracy scores of the individual models as variables
      accuracy = str(accuracy_score(y_test, prediction))

      ##making a list of individual models and their accuracy scores
      acc_score.append([alg.__class__.__name__, accuracy])

  ##creating a dataframe of every individual model and accuracy score 
  df = pd.DataFrame(acc_score, columns=['Model', 'Accuracy'])
  ##outputing the created dataframe 
  print(df)

  #return args

        ##creating another function for handling the categorial encoding of data 


def EncodeCategorical(Y): ##defining the argument the function takes
    from sklearn.preprocessing import LabelEncoder ##importing the needed library to handle data encoding
    le = LabelEncoder() ## assigning the encoder function to a variable
    return le.fit_transform(Y) ##returning the transformed encoded data 


def ml(X, Y, size=0.25, *args):
  return ML(X, Y, size=0.25, *args)


# -*- coding: utf-8 -*-
"""visualize.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1M7_YnsCAo2eJz_EDGOBffRNvqtmuY7gm
"""

def nn_visualize(model,type="keras",title ="My neural network", verbose =False):
  """
  nn_visualize is used to visualize neural networks in pictorial form(keras) or graphical form(graphviz)

  Parameters:
  model - reference of the model
  type - to return keras or graphviz visualization
  title - for graphviz visualization
  verbose(bool) - to display details of neural network architecture

  Return :
  Visualization object
  """



  if type == "keras" or type[0]=="k":
    from keras.utils.vis_utils import plot_model
    obj = plot_model(model, to_file='model_plot.png', show_shapes=True, show_layer_names=True)
    return obj

  else:
    try:
      from ann_visualizer.visualize import ann_viz
      from graphviz import Source

      ann_viz(model,view=True, title=title)
      graph_source = Source.from_file("network.gv")

      if verbose:
        print(graph_source.source)

      return graph_source
    except:
      print("An error occur while using graphviz visualization. So keras visualization object is returned")
      from keras.utils.vis_utils import plot_model
      obj = plot_model(model, to_file='model_plot.png', show_shapes=True, show_layer_names=True)
      return obj


def image_preprocess(dataframe):
    """
    used to preprocess image dataset to resize , rescale , convert to greyscale, normalisation to make it suitable for training

    Parameters:
    dataframe: image in dataframe

    Returns :
    preprocessed image dataset

    """

  pixels = df['pixels'].tolist() # Converting the relevant column element into a list for each row
  width, height = 48, 48
  images = []

  for pixel_sequence in pixels:
    img = [int(pixel) for pixel in pixel_sequence.split(' ')] # Splitting the string by space character as a list
    img = np.asarray(face).reshape(width, height) #converting the list to numpy array in size of 48*48
    img = cv2.resize(face.astype('uint8'),image_size) #resize the image to have 48 cols (width) and 48 rows (height)
    images.append(face.astype('float32')) #makes the list of each images of 48*48 and their pixels in numpyarray form
    
  images = np.asarray(images) #converting the list into numpy array
  images = np.expand_dims(images, -1) #Expand the shape of an array -1=last dimension => means color space


  x = images.astype('float32')
  x = x / 255.0 #Dividing the pixels by 255 for normalization  => range(0,1)

  # Scaling the pixels value in range(-1,1)
  x = x - 0.5
  x = x * 2.0

  return x





def testImage(imagePath) :

    """

    Used to preprocess test image - resize, reshape, greyscale , normalise to feed as input to model for predictions

    Parameters:
    imagepath(str): passes the path of the image

    Returns:
    prediction
    """



  import cv2
  import numpy as np
  
  image = cv2.imread(imagePath)
  gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
  cropped_img = np.expand_dims(np.expand_dims(cv2.resize(gray, (48, 48)), -1), 0)
  prediction = model.predict(cropped_img)

  maxindex = int(np.argmax(prediction))

  if img_show == True:
    cv2.imshow(image)
  return maxindex





def split(x ,y,verbose=True):

"""
Used to split dataset into training and validation dataset

Parameters:
x(list): the input dataset
y(list): output vector
verbose(bool): used to print information about dataset

Returns:
 tuple --> test and  validate dataset
 """
    import numpy
    import pandas

    num_samples, num_classes = y.shape

    num_samples = len(x)
    num_train_samples = int((1 - 0.2)*num_samples)

    # Traning data
    train_x = x[:num_train_samples]
    train_y = y[:num_train_samples]

    # Validation data
    val_x = x[num_train_samples:]
    val_y = y[num_train_samples:]

    train_data = (train_x, train_y)
    val_data = (val_x, val_y)

    if verbose==True:
        print('Training Pixels',train_x.shape)  # ==> 4 dims -  no of images , width , height , color
        print('Training labels',train_y.shape)

        print('Validation Pixels',val_x.shape)
        print('Validation labels',val_y.shape)

    return train_data,val_data


def model(input_shape=(48,48,1),arch ="simple_cnn",classify="Multi",num_classes=2):


  """
  Defines the models both architecture and configuration and returns the reference of the model 
  It consists of famous neural network architectures which helps us to implement these models in a single line.

  Parameters:
  input_shape(tuple): To define the input layer size.

  arch(str): To define the popular neural network architectures

  classify(str): It defines activation layer at outerlayer

  num_classes (int): To define number of neurons at output layer

  Returns:
  reference: The reference of the defined model.



  
  """




  from keras import layers, models,regularizers

    #load the libaray to built the model
  from keras.layers import Activation, Convolution2D, Dropout, Conv2D
  from keras.layers import AveragePooling2D, BatchNormalization
  from keras.layers import GlobalAveragePooling2D
  from keras.models import Sequential
  from keras.layers import Flatten
  from keras.models import Model
  from keras.layers import Input
  from keras.layers import MaxPooling2D
  from keras.layers import SeparableConv2D
  from keras import layers
  from keras.regularizers import l2
  from keras.utils import to_categorical
  from keras.models import Sequential
  from keras.layers import Dense, Dropout, Embedding, LSTM, GlobalMaxPooling1D, SpatialDropout1D

  if classify =="binary":
    classifier_function="sigmoid"
    loss_function = "binary_crossentropy"
  else:
    classifier_function="softmax"
    loss_function="categorical_crossentropy"

  if arch == "simple_cnn":
    #simple cnn
    model = Sequential()
    model.add(Convolution2D(filters=16, kernel_size=(7, 7), padding='same',
                                name='image_array', input_shape=input_shape))
    model.add(BatchNormalization())
    model.add(Convolution2D(filters=16, kernel_size=(7, 7), padding='same'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(AveragePooling2D(pool_size=(2, 2), padding='same'))
    model.add(Dropout(.5))

    model.add(Convolution2D(filters=32, kernel_size=(5, 5), padding='same'))
    model.add(BatchNormalization())
    model.add(Convolution2D(filters=32, kernel_size=(5, 5), padding='same'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(AveragePooling2D(pool_size=(2, 2), padding='same'))
    model.add(Dropout(.5))

    model.add(Convolution2D(filters=64, kernel_size=(3, 3), padding='same'))
    model.add(BatchNormalization())
    model.add(Convolution2D(filters=64, kernel_size=(3, 3), padding='same'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(AveragePooling2D(pool_size=(2, 2), padding='same'))
    model.add(Dropout(.5))

    model.add(Convolution2D(filters=128, kernel_size=(3, 3), padding='same'))
    model.add(BatchNormalization())
    model.add(Convolution2D(filters=128, kernel_size=(3, 3), padding='same'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(AveragePooling2D(pool_size=(2, 2), padding='same'))
    model.add(Dropout(.5))

    model.add(Convolution2D(filters=256, kernel_size=(3, 3), padding='same'))
    model.add(BatchNormalization())
    model.add(Convolution2D(filters=num_classes, kernel_size=(3, 3), padding='same'))
    model.add(GlobalAveragePooling2D())
    model.add(Activation(classifier_function,name='predictions'))

    model.compile('adam', loss_function, metrics=['acc'])

    return model



  elif arch in ["googleNet","inception"]:
    img_input = Input(input_shape)
    x = Conv2D(8, (3, 3), strides=(1, 1), kernel_regularizer=regularization,
                                                  use_bias=False)(img_input)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = Conv2D(8, (3, 3), strides=(1, 1), kernel_regularizer=regularization,
                                                  use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)

    # module 1
    residual = Conv2D(16, (1, 1), strides=(2, 2),padding='same', use_bias=False)(x)
    residual = BatchNormalization()(residual)

    x = SeparableConv2D(16, (3, 3), padding='same',
                            kernel_regularizer=regularization,
                            use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = SeparableConv2D(16, (3, 3), padding='same',
                            kernel_regularizer=regularization,
                            use_bias=False)(x)
    x = BatchNormalization()(x)

    x = MaxPooling2D((3, 3), strides=(2, 2), padding='same')(x)
    x = layers.add([x, residual])

    # module 2
    residual = Conv2D(32, (1, 1), strides=(2, 2),
                          padding='same', use_bias=False)(x)
    residual = BatchNormalization()(residual)

    x = SeparableConv2D(32, (3, 3), padding='same',
                            kernel_regularizer=regularization,
                            use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = SeparableConv2D(32, (3, 3), padding='same',
                            kernel_regularizer=regularization,
                            use_bias=False)(x)
    x = BatchNormalization()(x)

    x = MaxPooling2D((3, 3), strides=(2, 2), padding='same')(x)
    x = layers.add([x, residual])

    # module 3
    residual = Conv2D(64, (1, 1), strides=(2, 2),
                          padding='same', use_bias=False)(x)
    residual = BatchNormalization()(residual)

    x = SeparableConv2D(64, (3, 3), padding='same',
                            kernel_regularizer=regularization,
                            use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = SeparableConv2D(64, (3, 3), padding='same',
                            kernel_regularizer=regularization,
                            use_bias=False)(x)
    x = BatchNormalization()(x)

    x = MaxPooling2D((3, 3), strides=(2, 2), padding='same')(x)
    x = layers.add([x, residual])

    # module 4
    residual = Conv2D(128, (1, 1), strides=(2, 2),
                          padding='same', use_bias=False)(x)
    residual = BatchNormalization()(residual)

    x = SeparableConv2D(128, (3, 3), padding='same',
                            kernel_regularizer=regularization,
                            use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = SeparableConv2D(128, (3, 3), padding='same',
                            kernel_regularizer=regularization,
                            use_bias=False)(x)
    x = BatchNormalization()(x)

    x = MaxPooling2D((3, 3), strides=(2, 2), padding='same')(x)
    x = layers.add([x, residual])

    x = Conv2D(num_classes, (3, 3),
                #kernel_regularizer=regularization,
                padding='same')(x)
    x = GlobalAveragePooling2D()(x)
    output = Activation(classifier_function,name='predictions')(x)

    model = Model(img_input, output)

    model.compile('adam', loss_function, metrics=['acc'])

    return model


  elif arch == "basic_cnn" :#small cnn
    model = Sequential() 
    model.add(Conv2D(32, (2, 2), input_shape=input_shape)) 
    model.add(Activation('relu')) 
    model.add(MaxPooling2D(pool_size=(2, 2))) 
      
    model.add(Conv2D(32, (2, 2))) 
    model.add(Activation('relu')) 
    model.add(MaxPooling2D(pool_size=(2, 2))) 
      
    model.add(Conv2D(64, (2, 2))) 
    model.add(Activation('relu')) 
    model.add(MaxPooling2D(pool_size=(2, 2))) 
      
    model.add(Flatten()) 
    model.add(Dense(64)) 
    model.add(Activation('relu')) 
    model.add(Dropout(0.5)) 
    model.add(Dense(1)) 
    model.add(Activation(classifier_function)) 

    model.compile('adam', loss_function, metrics=['acc'])

    return model




  #Lenet
  elif arch=="lenet5":
    model = keras.Sequential()

    model.add(layers.Conv2D(filters=6, kernel_size=(3, 3), activation='relu', input_shape=input_shape))
    model.add(layers.AveragePooling2D())

    model.add(layers.Conv2D(filters=16, kernel_size=(3, 3), activation='relu'))
    model.add(layers.AveragePooling2D())

    model.add(layers.Flatten())

    model.add(layers.Dense(units=120, activation='relu'))

    model.add(layers.Dense(units=84, activation='relu'))

    model.add(layers.Dense(units=10, activation = classifier_function))

    model.compile('adam', loss_function, metrics=['acc'])
    
    return model

  #vgg16
  elif arch == "vgg16":
    model = Sequential()
    model.add(Conv2D(input_shape=input_shape,filters=64,kernel_size=(3,3),padding="same", activation="relu"))
    model.add(Conv2D(filters=64,kernel_size=(3,3),padding="same", activation="relu"))
    model.add(MaxPool2D(pool_size=(2,2),strides=(2,2)))
    model.add(Conv2D(filters=128, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(Conv2D(filters=128, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(MaxPool2D(pool_size=(2,2),strides=(2,2)))
    model.add(Conv2D(filters=256, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(Conv2D(filters=256, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(Conv2D(filters=256, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(MaxPool2D(pool_size=(2,2),strides=(2,2)))
    model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(MaxPool2D(pool_size=(2,2),strides=(2,2)))
    model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same", activation="relu"))
    model.add(MaxPool2D(pool_size=(2,2),strides=(2,2)))
    model.add(Flatten())
    model.add(Dense(units=4096,activation="relu"))
    model.add(Dense(units=4096,activation="relu"))
    model.add(Dense(units=2, activation="classifier_function"))

    model.compile('adam', loss_function, metrics=['acc'])

    return model

  #AlexNEt
  elif arch =="alexNet":
    #Instantiation
    AlexNet = Sequential()

    #1st Convolutional Layer
    AlexNet.add(Conv2D(filters=96, input_shape=input_shape, kernel_size=(11,11), strides=(4,4), padding='same'))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation('relu'))
    AlexNet.add(MaxPooling2D(pool_size=(2,2), strides=(2,2), padding='same'))

    #2nd Convolutional Layer
    AlexNet.add(Conv2D(filters=256, kernel_size=(5, 5), strides=(1,1), padding='same'))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation('relu'))
    AlexNet.add(MaxPooling2D(pool_size=(2,2), strides=(2,2), padding='same'))

    #3rd Convolutional Layer
    AlexNet.add(Conv2D(filters=384, kernel_size=(3,3), strides=(1,1), padding='same'))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation('relu'))

    #4th Convolutional Layer
    AlexNet.add(Conv2D(filters=384, kernel_size=(3,3), strides=(1,1), padding='same'))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation('relu'))

    #5th Convolutional Layer
    AlexNet.add(Conv2D(filters=256, kernel_size=(3,3), strides=(1,1), padding='same'))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation('relu'))
    AlexNet.add(MaxPooling2D(pool_size=(2,2), strides=(2,2), padding='same'))

    #Passing it to a Fully Connected layer
    AlexNet.add(Flatten())
    # 1st Fully Connected Layer
    AlexNet.add(Dense(4096, input_shape=(32,32,3,)))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation('relu'))
    # Add Dropout to prevent overfitting
    AlexNet.add(Dropout(0.4))

    #2nd Fully Connected Layer
    AlexNet.add(Dense(4096))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation('relu'))
    #Add Dropout
    AlexNet.add(Dropout(0.4))

    #3rd Fully Connected Layer
    AlexNet.add(Dense(1000))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation('relu'))
    #Add Dropout
    AlexNet.add(Dropout(0.4))

    #Output Layer
    AlexNet.add(Dense(10))
    AlexNet.add(BatchNormalization())
    AlexNet.add(Activation(classifier_function))

    AlexNet.compile('adam', loss_function, metrics=['acc'])
    return AlexNet



  #Basic MLP
  elif arch =="basic_mlp":
      # Create the model
    model = Sequential()
    model.add(Dense(350, input_shape=input_shape, activation='relu'))
    model.add(Dense(50, activation='relu'))
    model.add(Dense(num_classes, activation=classifier_function))

  elif arch == "deep_mlp":
    # Here's a Deep Dumb MLP (DDMLP)
    model = Sequential()
    model.add(Dense(128, input_dim=input_shape))
    model.add(Activation('relu'))
    model.add(Dropout(0.15))
    model.add(Dense(128))
    model.add(Activation('relu'))
    model.add(Dropout(0.15))
    model.add(Dense(nb_classes))
    model.add(Activation(classifier_function))

    model.compile('adam', loss_function, metrics=['acc'])

    return model

  #RNN
  elif arch == "basic_lstm":
    model = models.Sequential()
    model.add(layers.Embedding(10000, 128, input_length=40))
    model.add(layers.LSTM(32, activation='tanh'))
    model.add(layers.Dense(1, activation=classifier_function))

  #LSTM
  elif arch == "deep_lstm":
    model_lstm = Sequential()
    model_lstm.add(Embedding(input_dim = max_words, output_dim = 256, input_length = max_phrase_len))
    model_lstm.add(SpatialDropout1D(0.3))
    model_lstm.add(LSTM(256, dropout = 0.3, recurrent_dropout = 0.3))
    model_lstm.add(Dense(256, activation = 'relu'))
    model_lstm.add(Dropout(0.3))
    model_lstm.add(Dense(5, activation = classifier_function))



from setuptools import setup


def readme():
    with open('README.md') as f:
        README = f.read()
    return README

with open('requirements.txt') as f:
    required = f.read().splitlines()


setup(
    name="py-automl",
    version="1.0.6",
    description="py-automl - An open source, low-code machine learning library in Python.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/PrudhviGNV/py-automl",
    author="Prudhvi GNV",
    author_email="prudhvi.gnv@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=["pyAutoML"],
    include_package_data=True,
    install_requires=required

)
